#!/usr/bin/perl
use Apache::ConfigFile;
use strict;
use vars qw( %bin $dumpDir $ptrArgs $pwd $workDir $storage);

my $VERSION='1.0';

#
# parse command line
#
my %arg_opts = ('--help|-h'=>'',
				'--debug|-d'=>'i?',
				'--get-status|-s'=>'',
				'--dump-accounts|-dc'=>'s',
				'--dump-domains|-dd'=>'s',
				'--dump-all|-da'=>'',
				'--get-content-list|-lc'=>'',
				'--no-content|-nc'=>'',
				'--no-compress|-nz'=>'',
				'--output|-o'=>'s',
				'--status-file|-sf'=>'s',
				'--version|-v'=>'',
			   );

%bin = ('tar'=>'tar','bsdtar'=>'bsdtar','split'=>'split','cp'=>'cp','rm'=>'rm','echo'=>'echo',
		'mysql'=>'mysql','mysqldump'=>'mysqldump',
		'psql'=>'psql','uptime'=>'uptime','uname'=>'uname',
		'cat'=>'cat','pwd'=>'pwd','md5sum'=>'md5sum',
		'grep'=>'grep','find'=>'find','gzip'=>'gzip','wc'=>'wc','rpm'=>'rpm',
		'locale'=>'locale', 'iconv'=>'iconv',
		'python'=>'python','apachectl'=>'apachectl','apache2ctl'=>'apache2ctl');

my $returnCode = 0;

my $ptrHash = {};
my $ptrIpHash = {};

#
# global variables
#

my $kilobyte = 1024;
my $megabyte = $kilobyte*$kilobyte;

my $ignoredUsers_ = {'root' => '1',
					 'bin' => '1',
					 'daemon' => '1',
					 'adm' => '1',
					 'lp' => '1',
					 'sync' => '1',
					 'shutdown' => '1',
					 'halt' => '1',
					 'mail' => '1',
					 'uucp' => '1',
					 'operator' => '1',
					 'games' => '1',
					 'gopher' => '1',
					 'ftp' => '1',
					 'news' => '1',
					 'mysql' => '1',
					 'nobody' => '1',
					 'zope' => '1',
					 'majordomo' => '1',
					 'tomcat4' => '1',
					 'apache' => '1',
					 'ensimrootmail' => '1',
					 'sshd' => '1',
					};

my %ignoredUsers = %{$ignoredUsers_};

#
# end global variables
#

require 'agent.include.pl';

$ptrArgs = &getArguments (\@ARGV,\%arg_opts);

#
# search binaries
#

&getBins(\%bin,'/bin:/usr/bin:/usr/local/bin:/sbin:/usr/sbin:'.
		 '/usr/local/sbin:'.$ENV{'PATH'});

my ($outPath,$outFh,$outHas,$dumpFile,$statusFile,$app);

$app = $0;
if ($app=~/(.+)\/([^\/]+)$/) {
  $workDir = $1;
  $app = $2;
}
$app=~s/\.pl$//;

$pwd=`$bin{'pwd'}`||'.';

chop $pwd;

if ($workDir eq '.') {
  $workDir = $pwd;
} elsif ($workDir !~ /^\//) {
  $workDir = $pwd.'/'.$workDir;
}

$dumpFile = $workDir.'/dump.xml';
$statusFile = $workDir.'/dumping-status.xml';
$dumpDir = $workDir.'/pma';

my $objDumpStatus = &makeDumpStatus($ptrArgs->{'status-file'}||$statusFile);

if (exists $ptrArgs->{'version'}) {
  print $outFh $VERSION,"\n";
  exit 0;
}

my $rootName = 'migration-dump';
my $dtdName = 'plesk.dtd';


if (exists $ptrArgs->{'get-status'}) {
  my $outFh = &openOutput($ptrArgs->{'output'});

  &printAgentStatus($outPath=$ptrArgs->{'output'}||$dumpFile,
					$objDumpStatus->{'FILE'}->(), $outFh);

  &closeOutput();
  exit 0;
}

#
# get MIME Base64 encoder
#
my $wrapBase64 = makeMIMEBase64();


if (exists ($ptrArgs->{'dump-all'}) ||
	exists ($ptrArgs->{'dump-accounts'}) ||
	exists ($ptrArgs->{'dump-domains'})) {

  my ($root,@accounts,@domains,$ptrAccounts,$ptrDomains, $value);
  initStorage();
  getDumpDir($dumpDir);

  if (exists $ptrArgs->{'no-compress'}) {
	setCompress(0);
  }

  &printToLog("Work dir: $workDir");
  &printToLog("Dump file: $dumpFile");
  &printToLog("Status file: ".$objDumpStatus->{'FILE'}->());

 if ($value = $ptrArgs->{'dump-domains'}) {
	if ($value eq "-") {
	  $value = <STDIN>;
	  chomp $value;
	}
	@domains = split(/\s*,\s*/,$value);
	$ptrDomains = \@domains;
 }
  #
  # generate a xml dump
  #

  $root = &getAlienDump($dumpDir, $ptrDomains, $rootName);

  #
  # print dump to output
  #
  $storage->finish($root);
  &printToLog(' OK');

} elsif (exists $ptrArgs->{'get-content-list'}) {

  makeContentList();
  exit 0;

} else {
  &printHelp();
}

exit $returnCode;

#
# Begin main
#

sub printAgentStatus {
  my $dumpFile = shift;
  my $statusFile = shift;
  my $outFh = shift || *STDOUT;
  my $rootName = shift || 'agentstatus';

  my ($root,$out,$item);

  $root = &makeXmlNode($rootName);

  $item = &makeXmlNode('dumpfile',$dumpFile);
  $root->{'ADDCHILD'}->($item);
  $item = &makeXmlNode('statusfile',$statusFile);
  $root->{'ADDCHILD'}->($item);

  &printXml($root,$outFh, 'plesk.dtd');
}

sub detectSoft {

	my ($httpd,$apachectl);

	open (NS, "netstat -lp --numeric-port |");
        while(<NS>){
                if ($_=~ /.+\:80\b/) {
                 	if ($_=~/(\d+)\/(\w+)/){
                    		$httpd=$2;
                	}
        	}
	}

	close(NS);

	if ($httpd eq 'httpd2' || $httpd eq 'apache2'){
		$apachectl = $bin{'apache2ctl'} || 'apache2ctl';
	}else{
		$apachectl = $bin{'apachectl'} || 'apachectl';
	}

	my ($htRoot,$config,$configPath);
	open (CTL, "$apachectl -V|");
	while (my $line = <CTL>){
		chomp $line;
		if ($line =~/\w+\s+(HTTPD_ROOT)=(.+)/){
			$htRoot = $2;
			$htRoot =~s/"//g;
		}
		if ($line =~/\w+\s+(SERVER_CONFIG_FILE)=(.+)/){
			$config = $2;
			$config	=~s/"//g;
		}
	}
	return $htRoot."/".$config;
}

sub getAlienDump {
	my ($dumpDir, $ptrDomains,$rootName)=@_;

	my $root = makeXmlNode($rootName);
        $root->{'ATTRIBUTE'}->('agent-name','generic');

	my $httpd_conf = detectSoft();

	getAllIncludeFiles($httpd_conf);

	my $clientNode = createFakeClient($root);

	if (ref($ptrDomains)=~/ARRAY/ && (@{$ptrDomains} > 0)) {

		createSelectedDomains($clientNode,$ptrDomains);

	}else{
		createAllDomains($clientNode);
	}

	return $root;

}

sub createFakeClient {
	my ($parent) = @_;
	my $item = makeXmlNode('client');
	$item->{'ATTRIBUTE'}->('name','TheFakeClient');
	createPasswordNode($item,'','plain');
    my $statusNode = makeXmlNode('status');
    $statusNode->{'ADDCHILD'}->(makeXmlNode('enabled'));
    $item->{'ADDCHILD'}->($statusNode);
	my $ipPoolNode = makeXmlNode('ip_pool');

	createClientIpNode($ipPoolNode);

	$item->{'ADDCHILD'}->($ipPoolNode);
	$parent->{'ADDCHILD'}->($item);

	return $item;
}

sub createClientIpNode {
	my ($parent) = @_;
	foreach my $domain (keys %{$ptrIpHash}){
		my $clIpNode = makeXmlNode('ip');
		$clIpNode->{'ADDCHILD'}->(makeXmlNode('ip-type','shared'));
                $clIpNode->{'ADDCHILD'}->(makeXmlNode('ip-address',$ptrIpHash->{$domain}));
                $parent->{'ADDCHILD'}->($clIpNode);
	}
}

sub createSelectedDomains {
	my ($parent,$ptrDomains)=@_;
	my $domain;
	my @domains = @{$ptrDomains};

	foreach $domain (@domains){
		if (exists $ptrHash->{$domain}){
			createDomainNode($parent,$ptrHash->{$domain});
		}
	}
}

sub createAllDomains {
	my ($parent)=@_;
	my $domain;

	foreach $domain (keys %{$ptrHash}){
		createDomainNode($parent,$ptrHash->{$domain});
	}
}

sub createDomainNode {
  my ($parent,$objVh)=@_;
  my $item = makeXmlNode('domain');

  my ($serverAlias,$serverName,$sslObj);

  $serverAlias = $objVh->{'cmd_context'}->{'ServerAlias'}->[0]->[0];
  $serverName  = $objVh->{'cmd_context'}->{'ServerName'}->[0]->[0];

  $serverName =~s/:[0-9]+//g;

  createDomainIpNode($item,$ptrIpHash->{$serverName});

  my $statusNode = makeXmlNode('status');
  $statusNode->{'ADDCHILD'}->(makeXmlNode('enabled'));
  $item->{'ADDCHILD'}->($statusNode);

  $item->{'ATTRIBUTE'}->('name',$serverName);

  if(exists $ptrArgs->{'no-content'}){
    $parent->{'ADDCHILD'}->($item);
    return;
  }
  $item->{'ATTRIBUTE'}->('www', 'true') if defined($serverAlias);

  createPhostingNode($item,$objVh,$serverName,$sslObj);
  $parent->{'ADDCHILD'}->($item);
}

sub createDomainIpNode {
        my ($parent,$ipAddr) = @_;
        my $item = makeXmlNode('ip');
        $item->{'ADDCHILD'}->(makeXmlNode('ip-type','shared'));
        $item->{'ADDCHILD'}->(makeXmlNode('ip-address',$ipAddr));
        $parent->{'ADDCHILD'}->($item);
}

sub getAllIncludeFiles {
	my $config=shift;
        my @fileArray;
        my $file;
        local (*CONF);
        open(CONF,"< $config");
        while(<CONF>){
                next if ($_=~/^#/);
                if ($_=~/Include (.*)/){
                        push @fileArray,$1;
                }
        }
        close(CONF);

	foreach $file(@fileArray){
        	next if $file eq "conf.d/*.conf";
        	constructDump($file);
	        getAllIncludeFiles($file);
	}
}

sub constructDump {
	my ($pathConf)=@_;

	my $ac = Apache::ConfigFile->read($pathConf);
	my @vhkeys = $ac -> cmd_context('VirtualHost');
	my ($serverAlias,$serverName,$sslObj);

	for my $vhkey (@vhkeys) {
		my $vh = $ac -> cmd_context(VirtualHost => $vhkey);

		my $sname = $vh->cmd_config('ServerName');

		$sname=~s/:[0-9]+//g;

		$ptrHash->{$sname} = $vh;

		$vhkey=~s/:[0-9]+//g;

		$ptrIpHash->{$sname} = $vhkey;
	}

	return $ptrHash;
}

sub createPhostingNode {
	my ($parent,$objVh,$serverName,$objSsl)=@_;
	my $item = makeXmlNode('phosting');

	my ($cid,$cidCgi,$cidSsl);

# contents
	my $docRoot = $objVh->cmd_config('DocumentRoot');

	if (-d $docRoot){
		$cid = makeDumpFile("$dumpDir/$serverName\_docroot",$docRoot);
		$item->{'ATTRIBUTE'}->('cid_docroot',$cid);
	}

	my @scriptAlias = $objVh->cmd_config_array('ScriptAlias');

	for my $script (@scriptAlias){
		if ($script->[0]=~/cgi-bin/){
			$cidCgi = makeDumpFile("$dumpDir/$serverName\_cgi",$script->[1]);
			$item->{'ATTRIBUTE'}->('cid_cgi',$cidCgi);
		}
	}

	if (ref($objSsl) eq "ConfigFile"){
		my $docRootSsl = $objSsl->cmd_config('DocumentRoot');
		$cidSsl = makeDumpFile("$dumpDir/$serverName\_ssl",$docRootSsl);
		$item->{'ATTRIBUTE'}->('https','true');
		$item->{'ATTRIBUTE'}->('cid_docroot_ssl',$cidSsl);
	}
# sysuser

	createSysUserNode($item,$docRoot);

# scripting

	createScriptingNode($item,$objVh,$docRoot);

	$parent->{'ADDCHILD'}->($item);
}

sub createScriptingNode {
	my ($parent,$objVh,$docRoot)=@_;
	my $item = makeXmlNode('scripting');
	my $objScripts = $objVh->cmd_context(Directory => $docRoot);

#python
	if (ref($objScripts) eq "ConfigFile"){
		if(ref($objScripts->cmd_context(IfModule => 'mod_python.c')) eq "ConfigFile"){
			$item->{'ATTRIBUTE'}->('python','true');
		}

#cgi

	my @cgiAlias = $objVh->cmd_config_array('ScriptAlias');

	for my $script (@cgiAlias){
		if ($script->[0]=~/cgi-bin/){
			$item->{'ATTRIBUTE'}->('cgi','true');
		}
	}

#php

	my $objPhp = $objScripts->cmd_context(IfModule => 'mod_php5.c'); #need to think about other versions of mod_php.
	if (ref($objPhp) eq "ConfigFile") {
		my @phpOpts = $objPhp->cmd_config_array('php_admin_flag');

		for my $phpOpt (@phpOpts){
			if ($phpOpt->[1] eq "on"){
				$item->{'ATTRIBUTE'}->('php','true');
			}
		}
	}
#ssi
	my @opts = $objScripts->cmd_config_array('Options');

	for my $opt (@opts){
		if ($opt->[0] eq "+Includes"){
			$item->{'ATTRIBUTE'}->('ssi','true');
		}
	}

#mod_perl && asp

	my @modPerlOpts = $objScripts->cmd_context(IfModule => 'mod_perl.c');

	for my $modPerlOpt (@modPerlOpts){
		if (ref($modPerlOpt) eq "ConfigFile"){
			if ($modPerlOpt->cmd_config('Files') eq "\~ \(\\.asp\$\)") { $item->{'ATTRIBUTE'}->('asp','true')};
			if ($modPerlOpt->cmd_config('Files') eq "\~ \(\\.pl\$\)") { $item->{'ATTRIBUTE'}->('perl','true')};
		}
	}
	}

	$parent->{'ADDCHILD'}->($item);

}

sub createSysUserNode {
	my ($parent,$docRoot)=@_;
	my @dirProps = stat($docRoot);
	my $userParams = parsePasswdShadow(@dirProps->[4]);
	my $item = makeXmlNode('sysuser');
	my $username = $userParams->{@dirProps->[4]}->{'name'};
	if ($username eq 'root') {$username = 'toor'};

	$item->{'ATTRIBUTE'}->('name',$username);
	$item->{'ATTRIBUTE'}->('shell',$userParams->{@dirProps->[4]}->{'shell'});
	createPasswordNode($item,$userParams->{@dirProps->[4]}->{'password'},'encrypted');
	$parent->{'ADDCHILD'}->($item);
}

sub createPasswordNode {
	my ($parent, $password, $type) = @_;

	my $item = &makeXmlNode('password', $password);
	$item->{'ATTRIBUTE'}->('type', $type);
	$parent->{'ADDCHILD'}->($item);
}

sub parsePasswdShadow {

	my ($uid)=@_;

	my $passwdReader = makeFileParser("/etc/passwd");
	my $shadowReader = makeSafeFileParser("/etc/shadow");

	my (%passwd, %shadow);

        if (defined($passwdReader)) {
		%passwd = %{$passwdReader->{'PARSE'}->('KEYSEPARATOR' => ':', 'VALUESEPARATOR' => ':')};
	} else {
		return undef;
	}

	if (defined($shadowReader)) {
		%shadow = %{$shadowReader->{'PARSE'}->('KEYSEPARATOR' => ':',
										   'VALUESEPARATOR' => ':')};
  	}

	my ($user,$paramsPtr,@params,$password,$shell,@shadowParams,%userParams);

	while (($user,$paramsPtr) = each %passwd){
			my @params = @{$paramsPtr};
			if($uid == $params[1]){
				$shell = $params[5];

			if (%shadow && defined($shadow{$user})){
				@shadowParams = @{$shadow{$user}};
				$password = $shadowParams[0];
			}
		$userParams{$uid} = {'name' => $user,'shell' => $shell, 'password' => $password};
		}
	}

	return \%userParams;
}

sub printHelp {

  print <<"HELP";

Agent of migration from any platform to Plesk

Usage:
  $0 <options>

Options:
  -s |--get-status           get status of the agent

  -dd|--dump-domains=<list>  a comma separated list of domains to dump
  -da|--dump-all             make a full dump

  -lc|--get-content-list     get list of content files
  -nc|--no-content           do not make content files
  -nz|--no-compress          do not compress content files

  -o |--output=<path>        redirect output to file

  -h |--help                 this help
  -d |--debug=<level>        level of debug messages

HELP
}
